<?php

declare(strict_types=1);

namespace Drupal\rdf_sync;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\rdf_sync\Model\ColumnMapping;

/**
 * RDF Sync mapper service.
 */
class RdfSyncMapper {

  /**
   * Constructs a new sync mapper instance.
   *
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entityTypeBundleInfo
   *   The entity type bundle info service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   */
  public function __construct(
    protected EntityTypeBundleInfoInterface $entityTypeBundleInfo,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected Connection $database,
  ) {}

  /**
   * Returns the RDF type of given bundle.
   *
   * The caller should pass either $entityTypeId and $bundle or the $entity.
   *
   * @param string|null $entityTypeId
   *   The bundle's entity type ID.
   * @param string|null $bundle
   *   The bundle.
   * @param \Drupal\Core\Entity\ContentEntityInterface|null $entity
   *   The entity.
   */
  public function getBundleRdfType(?string $entityTypeId = NULL, ?string $bundle = NULL, ?ContentEntityInterface $entity = NULL): ?string {
    $this->resolveEntityTypeAndBundle($entityTypeId, $bundle, $entity);
    return $this->getBundleMapping($entityTypeId, $bundle)['type'] ?? NULL;
  }

  /**
   * Checks whether a given bundle is mapped.
   *
   * @param string $entityTypeId
   *   The bundle's entity type ID.
   * @param string $bundle
   *   The bundle.
   *
   * @return bool
   *   Whether a given bundle is mapped.
   */
  public function isMappedBundle(string $entityTypeId, string $bundle): bool {
    return (bool) $this->getBundleRdfType($entityTypeId, $bundle);
  }

  /**
   * Checks whether an entity is mapped.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return bool
   *   Whether the entity is mapped.
   */
  public function isMappedEntity(EntityInterface $entity): bool {
    return $this->isMappedBundle($entity->getEntityTypeId(), $entity->bundle());
  }

  /**
   * Returns the mappings of given bundle.
   *
   * The caller should pass either $entityTypeId and $bundle or the $entity.
   *
   * @param string|null $entityTypeId
   *   The bundle's entity type ID.
   * @param string|null $bundle
   *   The bundle.
   * @param \Drupal\Core\Entity\ContentEntityInterface|null $entity
   *   The entity.
   *
   * @return \Drupal\rdf_sync\Model\ColumnMapping[][]
   *   Mappings list.
   */
  public function getMappings(?string $entityTypeId = NULL, ?string $bundle = NULL, ?ContentEntityInterface $entity = NULL): array {
    $this->resolveEntityTypeAndBundle($entityTypeId, $bundle, $entity);

    $mappings = [];
    foreach ($this->getBundleMapping($entityTypeId, $bundle)['fields'] ?? [] as $fieldName => $columns) {
      foreach ($columns as $columnName => $mapping) {
        $mappings[$fieldName][$columnName] = ColumnMapping::createFromArray($mapping);
      }
    }

    return $mappings;
  }

  /**
   * Returns the URI field name.
   *
   * The caller should pass either $entityTypeId and $bundle or the $entity.
   *
   * @param string|null $entityTypeId
   *   The bundle's entity type ID.
   * @param string|null $bundle
   *   The bundle.
   * @param \Drupal\Core\Entity\ContentEntityInterface|null $entity
   *   The entity.
   *
   * @return string
   *   The URI field name.
   */
  public function getRdfUriFieldName(?string $entityTypeId = NULL, ?string $bundle = NULL, ?ContentEntityInterface $entity = NULL): string {
    $this->resolveEntityTypeAndBundle($entityTypeId, $bundle, $entity);
    return $this->getBundleMapping($entityTypeId, $bundle)['uri_field_name'];
  }

  /**
   * Returns the bundle RDF URI plugin ID.
   *
   * @param string $entityTypeId
   *   The bundle's entity type ID.
   * @param string $bundle
   *   The bundle.
   *
   * @return array|null
   *   The bundle RDF URI plugin ID.
   */
  public function getRdfUriPluginId(string $entityTypeId, string $bundle): ?string {
    return $this->getBundleMapping($entityTypeId, $bundle)['uri_plugin'] ?? NULL;
  }

  /**
   * Checks whether an entity did map the bundle.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity.
   *
   * @return bool
   *   Whether an entity did map the bundle.
   */
  public function hasBundleMapping(ContentEntityInterface $entity): bool {
    $bundleKey = $entity->getEntityType()->getKey('bundle');
    $fieldMappings = $this->getMappings($entity->getEntityTypeId(), $entity->bundle());
    if (!isset($fieldMappings[$bundleKey])) {
      return FALSE;
    }
    $mainProperty = $entity->getFieldDefinition($bundleKey)->getFieldStorageDefinition()->getMainPropertyName();
    return isset($fieldMappings[$bundleKey][$mainProperty]);
  }

  /**
   * Returns an entity given its URI.
   *
   * @param string $uri
   *   The entity URI.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface|null
   *   The entity or NULL if it cannot be retrieved.
   */
  public function getEntityByUri(string $uri): ?ContentEntityInterface {
    $row = $this->database->select('rdf_sync_uri')
      ->fields('rdf_sync_uri', ['entity_type', 'entity_id'])
      ->condition('uri', $uri)
      ->execute()
      ->fetch(\PDO::FETCH_NUM);
    if (!$row) {
      return NULL;
    }
    [$entityTypeId, $entityId] = $row;

    return $this->entityTypeManager->getStorage($entityTypeId)->load($entityId);
  }

  /**
   * Returns a list of URIs keyed by their entity ID.
   *
   * @param string $entityTypeId
   *   The entity type ID.
   * @param string|null $bundle
   *   (optional) If provided, the result will be narrowed to this bundle.
   *
   * @return array
   *   A list of URIs keyed by the entity ID.
   */
  public function getUrisByEntityType(string $entityTypeId, ?string $bundle = NULL): array {
    $query = $this->database->select('rdf_sync_uri')
      ->fields('rdf_sync_uri', ['entity_id', 'uri'])
      ->condition('entity_type', $entityTypeId);
    if ($bundle) {
      $query->condition('bundle', $bundle);
    }
    return $query->execute()->fetchAllKeyed();
  }

  /**
   * Returns the bundle mapping.
   *
   * @param string $entityTypeId
   *   The bundle's entity type ID.
   * @param string $bundle
   *   The bundle.
   *
   * @return array|null
   *   The bundle mapping or NULL if no mapping is defined.
   */
  protected function getBundleMapping(string $entityTypeId, string $bundle): ?array {
    // Cover entity types without bundles.
    $bundle = $bundle ?: $entityTypeId;
    $bundleInfo = $this->entityTypeBundleInfo->getBundleInfo($entityTypeId)[$bundle] ?? [];
    if (!$bundleInfo) {
      return NULL;
    }
    return $bundleInfo['rdf_sync'] ?? [];
  }

  /**
   * Resolves the entity type ID and the bundle, given the entity.
   *
   * @param string|null $entityTypeId
   *   The entity type ID passed by reference.
   * @param string|null $bundle
   *   The bundle passed by reference.
   * @param \Drupal\Core\Entity\ContentEntityInterface|null $entity
   *   The entity.
   */
  protected function resolveEntityTypeAndBundle(?string &$entityTypeId, ?string &$bundle, ?ContentEntityInterface $entity): void {
    if ($entity) {
      if ($entityTypeId || $bundle) {
        throw new \InvalidArgumentException('Cannot pass $entityTypeId or $bundle when $entity has been passed');
      }
      $entityTypeId = $entity->getEntityTypeId();
      $bundle = $entity->bundle();
    }
    elseif (!($entityTypeId && $bundle)) {
      throw new \InvalidArgumentException('Both, $entityTypeId and $bundle, should be passed when $entity is missed');
    }
  }

}
